#!/usr/bin/env python

"""
Moin wiki text serialiser.

Copyright (C) 2017, 2018, 2021, 2022, 2023 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from moinformat.serialisers.common import escape_attr, escape_text, Serialiser

class MoinSerialiser(Serialiser):

    "Serialisation of the page."

    input_formats = ["moin", "wiki"]
    formats = ["moin", "wiki"]

    # Node handler methods.

    def region(self, region):
        out = self.out

        if region.level:
            out(" " * region.indent + "{" * region.level)

            # Produce a header for regions within a top-level region.

            if region.type and region.type != "inline":

                # Obtain individual arguments, excluding the region type.

                args = region.args.split(" ")[1:]
                args_str = args and (" %s" % " ".join(args)) or ""

                out("#!%s%s\n" % (region.type, args_str))

        # Serialise the region content.

        self.visit_region(region)

        if region.level:
            out("%s%s" % ("}" * region.level, region.extra or ""))

    # Block node methods.

    def block(self, block):
        self.container(block)

    def defitem(self, defitem):
        self.out((defitem.extra and defitem.extra + "::" or "") + defitem.pad)
        self.container(defitem)

    def defterm(self, defterm):
        self.out(defterm.pad)
        self.container(defterm)
        self.out("::" + defterm.extra)

    def fontstyle(self, fontstyle):
        if fontstyle.emphasis:
            self.out("''")
        elif fontstyle.strong:
            self.out("'''")
        self.container(fontstyle)
        if fontstyle.emphasis:
            self.out("''")
        elif fontstyle.strong:
            self.out("'''")

    def heading(self, heading):
        self.out(heading.start_extra + "=" * heading.level + heading.start_pad)
        self.container(heading)
        self.out(heading.end_pad + "=" * heading.level + heading.end_extra)

    def larger(self, larger):
        self.out("~+")
        self.container(larger)
        self.out("+~")

    def list(self, list):
        self.container(list)

    def listitem(self, listitem):
        self.out("%s%s%s%s" % (
            listitem.indent * " ",
            listitem.marker,
            listitem.num and "#%s" % listitem.num or "",
            listitem.space))
        self.container(listitem)

    def macro(self, macro):

        # Special case of a deliberately unexpanded macro.

        if macro.nodes is None:
            return

        # Fallback case for when macros are not replaced.

        if not macro.nodes:
            self.out("<<%s%s>>" % (macro.name, macro.args and "(%s)" % ",".join(macro.args) or ""))

    def monospace(self, monospace):
        self.out("`")
        self.container(monospace)
        self.out("`")

    def smaller(self, smaller):
        self.out("~-")
        self.container(smaller)
        self.out("-~")

    def strikethrough(self, strikethrough):
        self.out("--(")
        self.container(strikethrough)
        self.out(")--")

    def subscript(self, subscript):
        self.out(",,")
        self.container(subscript)
        self.out(",,")

    def superscript(self, superscript):
        self.out("^")
        self.container(superscript)
        self.out("^")

    def table(self, table):
        self.container(table)

    def table_cell(self, table_cell):
        self.out("||")
        self.container(table_cell)

    def table_row(self, table_row):
        self.container(table_row)
        self.out("||")
        self.out(table_row.trailing)

    def underline(self, underline):
        self.out("__")
        self.container(underline)
        self.out("__")

    # Inline node methods.

    def anchor(self, anchor):
        self.out("((%s))" % anchor.target)

    def break_(self, break_):
        self.out("\n")

    def comment(self, comment):
        self.out("##%s%s" % (comment.comment, comment.extra))

    def directive(self, directive):
        self.out("#%s%s" % (directive.directive, directive.extra))

    def linebreak(self, linebreak):
        self.out(r"\\")

    def link(self, link):
        self.out("[[%s" % link.target)
        for node in link.nodes:
            self.out("|")
            node.visit(self)
        self.out("]]")

    def link_label(self, link_label):
        self.container(link_label)

    def link_parameter(self, link_parameter):
        s = link_parameter.text_content()
        key_value = s.split("=", 1)

        if len(key_value) == 1:
            self.out(key_value[0])
        else:
            self.out("=".join(key_value))

    def nbsp(self, nbsp):
        self.out(r"\_")

    def rule(self, rule):
        self.out("-" * (rule.height + 4))

    def table_attrs(self, table_attrs):
        self.out("<")
        self.container(table_attrs)
        if not table_attrs.incomplete:
            self.out(">")

    def table_attr(self, table_attr):
        if table_attr.concise:
            if table_attr.name == "bgcolor":
                self.out(table_attr.value)
            elif table_attr.name == "colspan":
                self.out("-%s" % table_attr.value)
            elif table_attr.name == "align":
                self.out(table_attr.value == "left" and "(" or table_attr.value == "right" and ")" or ":")
            elif table_attr.name == "rowspan":
                self.out("|%s" % table_attr.value)
            elif table_attr.name == "valign":
                self.out(table_attr.value == "top" and "^" or "v")
            elif table_attr.name == "width":
                self.out(table_attr.value)
        else:
            self.out("%s%s" % (
                escape_text(table_attr.name),
                table_attr.value is not None and "=%s%s%s" % (
                    table_attr.quote or '"',
                    escape_attr(table_attr.value),
                    table_attr.quote or '"')
                    or ""))

    def text(self, text):
        self.out(text.s)

    def transclusion(self, transclusion):
        self.out("{{%s" % transclusion.target)
        for node in transclusion.nodes:
            self.out("|")
            node.visit(self)
        self.out("}}")

    def verbatim(self, verbatim):
        self.out("<<<")
        self.out(verbatim.text)
        self.out(">>>")

serialiser = MoinSerialiser

# vim: tabstop=4 expandtab shiftwidth=4
