#!/usr/bin/env python

"""
Lexical partitioning of HTML document content.

Copyright (C) 2023 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# Lexical analysis state transition handler functions.

def tag_or_similar(text, pos):

    # Consult the text positions following the position indicated.

    if text[pos:pos+2] == "<!":
        if text[pos+3:pos+4] == "[":
            return IN_INCLUSION
        elif text[pos+3:pos+5] == "--":
            return IN_COMMENT
        else:
            return IN_DIRECTIVE
    else:
        return IN_TAG

def at_attribute_value(text, pos):
    return AT_ATTRIBUTE_VALUE

def in_dq_attribute_value(text, pos):
    return IN_DQ_ATTRIBUTE_VALUE

def in_sq_attribute_value(text, pos):
    return IN_SQ_ATTRIBUTE_VALUE

def after_attribute_value(text, pos):
    return AFTER_ATTRIBUTE_VALUE

def end_of_standalone_tag(text, pos):
    return AT_END_OF_TAG

def end_of_tag(text, pos):
    return BETWEEN_TAGS



# Lexical analysis states/spans.

class Span:
    def __init__(self, text):
        self.text = text

    def empty(self):
        return not self.text

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.text)

class AT_END_OF_TAG(Span):
    transitions = [(None, "", end_of_tag)]

    def empty(self):
        return False

    def visit(self, visitor):
        return visitor.at_end_of_tag(self)

class BETWEEN_TAGS(Span):
    transitions = [("<", "", tag_or_similar)]

    def visit(self, visitor):
        return visitor.between_tags(self)

class IN_TAG(Span):
    transitions = [
        ("=", "", at_attribute_value),
        ("/>", "", end_of_standalone_tag),
        (">", "", end_of_tag),
        ]

    def visit(self, visitor):
        return visitor.in_tag(self)

class IN_COMMENT(Span):
    transitions = [("-->", "--", end_of_tag)]

    def visit(self, visitor):
        return visitor.in_comment(self)

class IN_DIRECTIVE(Span):
    transitions = [(">", "", end_of_tag)]

    def visit(self, visitor):
        return visitor.in_directive(self)

class IN_INCLUSION(Span):
    transitions = [("]]>", "]]", end_of_tag)]

    def visit(self, visitor):
        return visitor.in_inclusion(self)

class AFTER_ATTRIBUTE_VALUE(Span):
    transitions = [
        ("=", "", at_attribute_value),
        ("/>", "", end_of_standalone_tag),
        (">", "", end_of_tag),
        ]

    def empty(self):
        return not self.text.strip()

    def visit(self, visitor):
        return visitor.after_attribute_value(self)

class AT_ATTRIBUTE_VALUE(Span):
    transitions = [
        ("=", "", at_attribute_value),
        ('"', "", in_dq_attribute_value),
        ("'", "", in_sq_attribute_value),
        ("/>", "", end_of_standalone_tag),
        (">", "", end_of_tag),
        ]

    def empty(self):
        return not self.text.strip()

    def visit(self, visitor):
        return visitor.at_attribute_value(self)

class IN_DQ_ATTRIBUTE_VALUE(Span):
    transitions = [('"', "", after_attribute_value)]

    def visit(self, visitor):
        return visitor.in_dq_attribute_value(self)

class IN_SQ_ATTRIBUTE_VALUE(Span):
    transitions = [("'", "", after_attribute_value)]

    def visit(self, visitor):
        return visitor.in_sq_attribute_value(self)



# Utility functions.

def find_one(text, pos, choices):

    """
    Find in 'text' from 'pos' the earliest occurring instance of one of the
    given 'choices', these being a list of (token string, extra string, state)
    tuples.

    The token string is a token marking the start of the next span, the extra
    string is the portion of the token to be added to the end of the current
    span upon matching, and the state applies to the next span.

    The associated state, the position of the occurrence, and the position of
    the text following the occurrence are returned as a tuple.
    """

    next_state = None
    first_pos = None
    first_extra = None
    next_pos = None

    for token, extra, state in choices:
        if token is None:
            return state, pos, extra, pos

        found_pos = text.find(token, pos)

        if found_pos != -1 and (next_state is None or found_pos < first_pos):
            next_state = state
            first_pos = found_pos
            first_extra = extra
            next_pos = found_pos + len(token)

    return next_state, first_pos, first_extra, next_pos



# Lexical partitioning.

class Lexer:
    def __init__(self, text):
        self.text = text
        self.state = BETWEEN_TAGS
        self.pos = 0

    def _end_of_input(self):
        start = self.pos
        self.pos = None
        return self._span(self.text[start:])

    def _span(self, text):
        return self.state(text)

    def __iter__(self):
        return self

    def next(self):
        if self.pos is None:
            raise StopIteration

        # Obtain details of a state transition: a handler function to determine
        # the next state, and the start and end positions of the token causing
        # the transition.

        handler, pos, extra, next_pos = find_one(self.text, self.pos, self.state.transitions)

        if handler is None:
            return self._end_of_input()

        # Obtain the lexical span and update the state and position.

        span = self._span(self.text[self.pos:pos] + extra)

        self.state = handler(self.text, pos)
        self.pos = next_pos

        return span

# vim: tabstop=4 expandtab shiftwidth=4
