#!/usr/bin/env python

"""
An absurdly minimal HTML tokeniser.

Copyright (C) 2023 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from collections import deque
from moinformat.utils.htmlparse.lex import Lexer



# Document token classes.

class Token:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.value)

class Attribute(Token):
    def visit(self, visitor):
        return visitor.attribute(self)

class AttributeValue(Token):
    def __init__(self, value, quote):
        self.value = value
        self.quote = quote

    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.value, self.quote)

    def visit(self, visitor):
        return visitor.attribute_value(self)

class Comment(Token):
    def visit(self, visitor):
        return visitor.comment(self)

class Directive(Token):
    def visit(self, visitor):
        return visitor.directive(self)

class Inclusion(Token):
    def visit(self, visitor):
        return visitor.inclusion(self)

class Tag(Token):
    def visit(self, visitor):
        return visitor.tag(self)

    def is_end(self):
        return self.value.startswith("/")

    def tag_name(self):
        return self.is_end() and self.value[1:] or self.value

class TagClose:
    def visit(self, visitor):
        return visitor.tag_close(self)

    def __repr__(self):
        return "%s()" % self.__class__.__name__

class Text(Token):
    def visit(self, visitor):
        return visitor.text(self)



# Tidying visitor employing the spans from lexical partitioning.

class Visitor:
    def __init__(self):
        self.queued = deque()

    def visit(self, span):
        return span.visit(self)

    # Specific handler methods.

    def between_tags(self, span):
        return Text(span.text)

    def in_comment(self, span):
        return Comment(span.text)

    def in_directive(self, span):
        return Directive(span.text)

    def in_inclusion(self, span):
        return Inclusion(span.text)

    def _queue_attributes(self, tokens):
        for token in tokens:
            self.queued.append(Attribute(token))

    def in_tag(self, span):
        tokens = span.text.split()
        self._queue_attributes(tokens[1:])
        return Tag(tokens[0])

    def at_end_of_tag(self, span):
        return TagClose()

    def after_attribute_value(self, span):
        tokens = span.text.split()
        self._queue_attributes(tokens)
        return self.queued.popleft()

    def at_attribute_value(self, span):
        tokens = span.text.split()
        self._queue_attributes(tokens[1:])
        return AttributeValue(tokens[0], "")

    def in_dq_attribute_value(self, span):
        return AttributeValue(span.text, '"')

    def in_sq_attribute_value(self, span):
        return AttributeValue(span.text, "'")



# Tokenising.

class Tokeniser:
    def __init__(self, text):
        self.lexer = Lexer(text)
        self.visitor = Visitor()

    def __iter__(self):
        return self

    def next(self):
        if self.visitor.queued:
            return self.visitor.queued.popleft()

        while 1:
            span = self.lexer.next()
            if not span.empty():
                break

        return self.visitor.visit(span)

# vim: tabstop=4 expandtab shiftwidth=4
